/*  "ents.scm" WB-tree File Based Associative String Data Base System. */
/*  Copyright (C) 1991, 1992, 1993, 2000, 2003, 2006 Free Software Foundation, Inc. */
/*  */
/*  This program is free software; you can redistribute it and/or modify */
/*  it under the terms of the GNU General Public License as published by */
/*  the Free Software Foundation; either version 2, or (at your option) */
/*  any later version. */
/*  */
/*  This program is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/*  GNU General Public License for more details. */
/*  */
/*  You should have received a copy of the GNU General Public License */
/*  along with this software; see the file COPYING.  If not, write to */
/*  the Free Software Foundation, 59 Temple Place, Suite 330, Boston, */
/*  MA 02111, USA. */
/*  */
/*  As a special exception, the Free Software Foundation gives permission */
/*  for additional uses of the text contained in its release of WB. */
/*  */
/*  The exception is that, if you link the WB library with other files */
/*  to produce an executable, this does not by itself cause the */
/*  resulting executable to be covered by the GNU General Public License. */
/*  Your use of that executable is in no way restricted on account of */
/*  linking the WB library code into it. */
/*  */
/*  This exception does not however invalidate any other reasons why */
/*  the executable file might be covered by the GNU General Public License. */
/*  */
/*  This exception applies only to the code released by the */
/*  Free Software Foundation under the name WB.  If you copy */
/*  code from other Free Software Foundation releases into a copy of */
/*  WB, as the General Public License permits, the exception does */
/*  not apply to the code that you add in this way.  To avoid misleading */
/*  anyone as to the status of such modified files, you must delete */
/*  this exception notice from them. */
/*  */
/*  If you write modifications of your own for WB, it is your choice */
/*  whether to permit this exception to apply to your modifications. */
/*  If you do not wish that, delete this exception notice. */



package wb;

import static wb.Scan.*;
import static wb.Blkio.*;
import static wb.Prev.*;
import static wb.Stats.*;
import static wb.Blink.*;
import static wb.Del.*;
import static wb.Segs.*;
import static wb.Handle.*;
import static wb.Wbdefs.*;
import static wb.BooleanMethods.*;
import static wb.Wbsys.*;
import static wb.Blk.*;
import static wb.Pkt.*;
import static wb.Ent.*;
import static wb.Seg.*;
import static wb.Han.*;
import static wb.Db.*;
import static wb.Lck.*;
public class Ents {

/*  tables */


public static Lck []lckTab = null;

public static Ent []bukTab = null;

public static Ent []entTab = null;

public static int numEntsCt = 0;

public static int numBuks = 0;

public static int blkSize = 0;

public static byte []emptyBlk = null;

public static Lck emptyBlkLck = null;


public static boolean cacheEntEnable_P = (true);

/*  DATABASE LEVEL OPERATIONS */

/*  This can be bummed to write less than the full BSIZ if we know */
/*  what the disk sector size is. */
/*  fixed order check in ent-write */


public static int ents_EntWrite(Ent ent)
{
  int seg = ent_Seg(ent);
  byte []blk = ent_Blk(ent);
  if (!(blk_Typ_P(blk, seqTyp)))
    checkKeyOrder(blk);
  blk_SetTime(blk, time(longZero));
  if ( -2==(seg_FlcLen(seg)))
    {
      dprintf(">>>>ERROR<<<< "+("entWrite")+" on read only segment "+(seg)+"?\n");
      return 0;
    }
  else if (a2b(blkio_Write(seg_Port(seg), ent_Blk(ent), seg_Bsiz(seg), ent_Id(ent))))
    {
      ent_SetDty(ent, null);
      return 1;
    }
  else {
    ent_SetDty(ent, true);
    return 0;
  }
}

/*  FLUSHING needs to be proportional to time (to put some limit on */
/*  how long things are left unwritten) plus write-activity. */

/*  NOTE: While flushing a buffer, get ACCPEND accmode to it (to */
/*  prevent surprise mods) */


public static int flushEntCntr = 0;

public static Lck flushEntLck = null;


public static int entsFlush(int trynum,int flushnum)
{
  int flushednum = 0;
  if (a2b(flushEntLck))
    if (tryLck_P(flushEntLck))
      {
	int iCt = trynum;
	while (!((0==(iCt)
		 || (flushednum)>=(flushnum)))) {
	  flushEntCntr = (1+(flushEntCntr))%(numEntsCt);
	  {
	    Ent tent = entTab[flushEntCntr];
	    if (ent_Dty_P(tent)
		    && !(a2b(ent_Acc(tent)))
	        && entFlush_P(tent))
	      {
		showBuffer1(tent);
		flushednum = (flushednum)+1;
	      }
	  }
	  {
	    iCt = (iCt)-1;
	  }
	}
	unlck(flushEntLck);
	return flushednum;
      }
    else return 0;
  else return 0;
}


public static boolean entFlush_P(Ent tent)
{
  int eSeg = ent_Seg(tent);
  int eId = ent_Id(tent);
  getBukLck_P(eSeg, eId);
  {
    Ent ent = getBuk(eSeg, eId);
Lloop: while (true) {
    if (!(a2b(ent)))
      {
	relBuk(eSeg, eId);
	return false;
      }
    else if ((tent)!=(ent))
      {
	ent = ent_Next(ent);
	continue Lloop;
      }
    else if (!(ent_Dty_P(ent)
	    && (accnone)==(ent_Acc(ent))))
      {
	relBuk(eSeg, eId);
	return false;
      }
    else {
      ent_SetAcc(ent, accpend);
      relBuk(eSeg, eId);
      ents_EntWrite(ent);
      getBukWait(eSeg, eId);
      ent_SetAcc(ent, accnone);
      relBuk(eSeg, eId);
      flushCt = (flushCt)+1;
      return (true);
    }
    }
  }
}

/*  release-ent! gives up all claim to ent, which is expected to be of */
/*  type accmode */
/*  fixed warning about dirty dirs -- twice */
/*  fixed dirty-block writer in UPDATE-ACCESS! */


public static void releaseEnt(Ent ent,int accmode)
{
  int blknum = ent_Id(ent);
  int seg = ent_Seg(ent);
  Ent buk = null;
  buk = getBukWait(seg, blknum);
  if ((accnone)!=(accmode)
      && (ent_Acc(ent))!=(accmode))
    dprintf(">>>>ERROR<<<< "+("releaseEnt")+": unexpected accmode of "+(seg)+":"+(blknum)+" is "+(ent_Acc(ent))
   +" not "+(accmode)+"\n");
  if ((accnone)==(accmode))
    ;
  else if (!(ent_Dty_P(ent)))
    ;
  else if (blk_Typ_P(ent_Blk(ent), dirTyp)
      && leaf_P(ent_Blk(ent)))
    {
      dprintf("WARNING: Directory block "+(seg)+":"+(blknum)+" dirty at RELEASE-ENT! \n");
      dirDtyCt = 1+(dirDtyCt);
    }
  else if (blk_Typ_P(ent_Blk(ent), seqTyp))
    {
      relBuk(seg, blknum);
      ents_EntWrite(ent);
      buk = getBukWait(seg, blknum);
    }
  if ((accnone)!=(accmode))
    ent_SetAcc(ent, accnone);
  if ((ent_Ref(ent))<=0)
    {
      ent_SetRef(ent, 0);
      dprintf(">>>>ERROR<<<< REF count below 0 in "+(seg)+":"+(blknum)+"\n");
    }
  else ent_SetRef(ent, (ent_Ref(ent))-1);
  if (0 > (seg))
    {
      spliceOutEnt(seg, blknum, buk, ent);
    }
  else ent_SetAge(ent, (ent_Dty_P(ent)
       ?5
       :0)+(5*(6+((leaf)-(blk_Level(ent_Blk(ent)))))));
  relBuk(seg, blknum);
}


public static boolean ents_EntUpdateAccess_P(Ent ent,int oldAccmode,int newAccmode)
{
  getBukWait(ent_Seg(ent), ent_Id(ent));
  if ((ent_Acc(ent))!=(oldAccmode))
    {
      relBuk(ent_Seg(ent), ent_Id(ent));
      dprintf(">>>>ERROR<<<< unexpected accmode type on "+(ent_Seg(ent))+":"+(ent_Id(ent))+" "+(ent_Acc(ent))+" (expected "+(oldAccmode)
   +")\n");
    }
  if ((accnone)==(oldAccmode))
    ;
  else if (!(ent_Dty_P(ent)))
    ;
  else if (blk_Typ_P(ent_Blk(ent), seqTyp))
    {
      relBuk(ent_Seg(ent), ent_Id(ent));
      ents_EntWrite(ent);
      getBukWait(ent_Seg(ent), ent_Id(ent));
    }
  else if (blk_Typ_P(ent_Blk(ent), dirTyp)
      && leaf_P(ent_Blk(ent)))
    {
      dprintf("WARNING: Directory block "+(ent_Seg(ent))+":"+(ent_Id(ent))+" dirty at ENT-UPD-ACCESS! \n");
      dirDtyCt = 1+(dirDtyCt);
    }
  ent_SetAcc(ent, newAccmode);
  relBuk(ent_Seg(ent), ent_Id(ent));
  return a2b(ent)
  && (true);
}

/*  ENT-FREE-LIST stuff ----------------------------------------------------------- */


public static int freeBukCntr = 0;

public static Lck freeEntLck = null;

public static Ent freeEnts = null;


public static Ent getFreeFreeEnt()
{
  lck(freeEntLck);
  if (a2b(freeEnts))
    {
      Ent freeEnt = freeEnts;
      freeEnts = ent_Next(freeEnts);
      unlck(freeEntLck);
      return freeEnt;
    }
  else return null;
}

/*  this version assumes the caller has already locked the bucket */
/*  BUK containing ENT */


public static void spliceOutEnt(int seg,int blkNum,Ent buk,Ent ent)
{
  {
    Ent bent = buk;
    Ent lastent = null;
    while (!((!(a2b(bent))
	     || (bent)==(ent)))) {
      {
	Ent T_bent = ent_Next(bent);
	lastent = bent;
	bent = T_bent;
      }
    }
    if (a2b(bent))
      {
	if (a2b(lastent))
	  {
	    ent_SetNext(lastent, ent_Next(bent));
	  }
	else setBuk(seg, blkNum, ent_Next(bent));
	recycleEnt(bent);
      }
    else dprintf("WARNING: "+("spliceOutEnt")+" couldn\'t "+(seg)+":"+(blkNum)+"\n");
  }
}


public static void recycleEnt(Ent ent)
{
  ent_SetDty(ent, null);
  ent_SetPus(ent, 0);
  ent_SetSeg(ent,  -1);
  ent_SetId(ent,  -1);
  lck(freeEntLck);
  ent_SetRef(ent, 0);
  ent_SetAcc(ent, accnone);
  ent_SetNext(ent, freeEnts);
  freeEnts = ent;
  unlck(freeEntLck);
}

/*  SELECT-IDLE-ENT selects a candidate entry for reuse.  caller needs to call */
/*  RECLAIM-ENT next to splice entry out of its bucket. */
/*  NOTE: when called, bucket (lseg lblk-num) is lcked. */
/*  The target bucket is assumed unlocked if lseg < 0. */
/*  (GET-ENT calls this with the bucket locked to prevent someone else from */
/*  getting another entry for the same block.) */


public static Ent selectIdleEnt(int lseg,int lblkNum)
{
  {
    Ent oldestEnt = null;
    int numScan = max(min(numBuks, 0xa), (numBuks)/0x14);
    int freeBase = freeBukCntr;
    freeBukCntr = ((numScan)+(freeBukCntr))%(numBuks);
    unlck(freeEntLck);
    {
      int i = 0;
      while (!(((i)>(numScan)
		   && a2b(oldestEnt)
	       || (i)>(numBuks)))) {
	{
	  int freeNum = ((freeBase)+(i))%(numBuks);
	  boolean dontLock_P = 0 > (lseg)
	    ?false
	    :(freeNum)==(hash2int(lseg, lblkNum));
	  if ((dontLock_P
		  || getBukLck_P(0, freeNum)))
	    {
	      Ent ent = getBuk(0, freeNum);
	      while (a2b(ent)) {
		if (0==(ent_Ref(ent)))
		  {
		    ent_SetAge(ent, (ent_Dty_P(ent)
			 ?1
			 :2)+(ent_Age(ent)));
		    if ((accnone)==(ent_Acc(ent))
		        && (!(a2b(oldestEnt))
			       || (ent_Age(ent))>(ent_Age(oldestEnt))))
		      oldestEnt = ent;
		  }
		{
		  ent = ent_Next(ent);
		}
	      }
	      if (dontLock_P)
		;
	      else relBuk(0, freeNum);
	    }
	}
	{
	  i = (i)+1;
	}
      }
      if ((i)>(numBuks))
	dprintf(">>>>ERROR<<<< No free ents\n");
      return oldestEnt;
    }
  }
}

/*  RECLAIM-ENT unlinks ENT from its bucket if its not in use. */
/*  It writes out the entry-s block if it's dirty */
/*  RECLAIM-ENT has 3 cases */
/*   (a) ENT is in use -- LSEG is unlocked, NIL is returned */
/*   (b) ENT is clean -- ENT is unlinked and returned */
/*   (c) ENT is DIRTY -- ENT is written, unlinked, and reclaimed (put on */
/*             free lsit); LSEG is UNLOCKED, NIL is returned. */
/*     possible optimization in case (c): if LSEG = -, */
/*     ENT could be written, unlinked, and returned (like (b)) */


public static Ent reclaimEnt(Ent ent,int lseg,int lblkNum)
{
  {
    int seg = ent_Seg(ent);
    int blkNum = ent_Id(ent);
    boolean segsEqual_P = !(0 > (lseg))
    && sameBuk_P(lseg, lblkNum, seg, blkNum);
    Ent buk = segsEqual_P
      ?getBuk(seg, blkNum)
      :getBukWait(seg, blkNum);
    if ((!(0==(ent_Ref(ent)))
        || (accnone)!=(ent_Acc(ent))))
      {
	relBuk(seg, blkNum);
	if (segsEqual_P)
	  ;
	else if (0 > (lseg))
	  ;
	else relBuk(lseg, lblkNum);
	dprintf("WARNING: "+("reclaimEnt")+": couldn\'t splice-out-ent "+(lseg)+":"+(lblkNum)+"\n");
	return null;
      }
    else {
      Ent bent = buk;
      Ent lastent = null;
      while (!((!(a2b(bent))
	       || (ent)==(bent)))) {
	{
	  Ent T_bent = ent_Next(bent);
	  lastent = bent;
	  bent = T_bent;
	}
      }
      if (!(a2b(bent)))
	{
	  relBuk(seg, blkNum);
	  if (segsEqual_P)
	    ;
	  else if (0 > (lseg))
	    ;
	  else relBuk(lseg, lblkNum);
	  dprintf(">>>>ERROR<<<< "+("reclaimEnt")+": couldn\'t find ent in bucket "+(seg)+":"+(blkNum)+" l="+(lseg)
   +":"+(lblkNum)+"\n");
	  return null;
	}
      else if (!(ent_Dty_P(ent)))
	{
	  if (a2b(lastent))
	    {
	      ent_SetNext(lastent, ent_Next(ent));
	    }
	  else setBuk(seg, blkNum, ent_Next(ent));
	  ent_SetNext(ent, null);
	  if (segsEqual_P)
	    ;
	  else relBuk(seg, blkNum);
	  return ent;
	}
      else {
	ent_SetAcc(ent, accpend);
	relBuk(seg, blkNum);
	if (segsEqual_P)
	  ;
	else if (0 > (lseg))
	  ;
	else relBuk(lseg, lblkNum);
	ents_EntWrite(ent);
	buk = getBukWait(seg, blkNum);
	ent_SetAcc(ent, accnone);
	spliceOutEnt(seg, blkNum, buk, ent);
	relBuk(seg, blkNum);
	return null;
      }
    }
  }
}

/*  TRY-GET-FREE-ENT either returns a free ent OR unlocks (lseg lblk-num) */


public static Ent tryGetFreeEnt(int lseg,int lblkNum)
{
  Ent ent = getFreeFreeEnt();
  if (!(a2b(ent)))
    {
      ent = selectIdleEnt(lseg, lblkNum);
      if (a2b(ent))
	{
	  ent = reclaimEnt(ent, lseg, lblkNum);
	}
      else if (0 > (lseg))
	;
      else relBuk(lseg, lblkNum);
    }
  return ent;
}


/*  Special entry points for Jonathan to do non-B-tree stuff. */
/*  Also now used in chain-scan. */


public static Ent allocateEnt()
{
  {
    Ent ent = tryGetFreeEnt( -1,  -1);
    int cnt = 0x13;
Lloop: while (true) {
    if (a2b(ent))
      {
	ent_SetAcc(ent, accwrite);
	ent_SetDty(ent, true);
	ent_SetPus(ent, 0);
	ent_SetSeg(ent,  -1);
	ent_SetId(ent,  -1);
	ent_SetRef(ent, 1);
	ent_SetNext(ent, null);
	return ent;
      }
    else if (0==(cnt))
      return null;
    else {
      ent = tryGetFreeEnt( -1,  -1);
      cnt =  -1+(cnt);
      continue Lloop;
    }
    }
  }
}


public static void entCopy(Ent toEnt,Ent fromEnt)
{
  if ((ent_Acc(toEnt))!=(accwrite))
    dprintf(">>>>ERROR<<<< "+("entCopy")+": copying into non-ACCWRITE "+(ent_Seg(toEnt))+":"+(ent_Id(toEnt))+"\n");
  ent_SetSeg(toEnt, ent_Seg(fromEnt));
  ent_SetId(toEnt, ent_Id(fromEnt));
  substringMove(ent_Blk(fromEnt), 0, seg_Bsiz(ent_Seg(fromEnt)), ent_Blk(toEnt), 0);
}


public static int getEntCopy(Ent toEnt,int seg,int blkNum)
{
  Ent fromEnt = getEnt(seg, blkNum, accread);
  if (a2b(fromEnt))
    {
      entCopy(toEnt, fromEnt);
      releaseEnt(fromEnt, accread);
      return 1;
    }
  else return 0;
}


public static void flushFlc(int seg,int fullness)
{
LflushFlc:while (true) {
  {
    byte []fstr = new byte[4];
    byte []tstr = new byte[4];
    lck(seg_Lck(seg));
    if ((seg_FlcLen(seg))<=(fullness))
      {
	unlck(seg_Lck(seg));
      }
    else {
      long2str(fstr, 0, seg_Flc(seg)[(seg_FlcLen(seg))-1]);
      seg_SetFlcLen(seg, (seg_FlcLen(seg))-1);
      unlck(seg_Lck(seg));
      long2str(tstr, 0, time(longZero));
      btPut(seg_FlHan(seg), fstr, 4, tstr, 4);
      continue LflushFlc;
    }
  }
  return;
}
}

/*  Assumes that SEG-LCK is locked by this process */


public static boolean initloadFlc_P(int seg)
{
  switch (seg_FlcLen(seg)) {
  case  -1:
    {
      byte []tmpStr = new byte[0x14];
      int flcImageLen = btGet(seg_RtHan(seg), flcByts, 3, tmpStr);
      if (0 > (flcImageLen))
	flcImageLen = 0;
      btPut(seg_RtHan(seg), flcByts, 3, noByts, 0);
      seg_SetFlcLen(seg, (flcImageLen)/4);
      {
	int i =  -4+(flcImageLen);
	while (!(0 > (i))) {
	  seg_Flc(seg)[(i)/4] = str2long(tmpStr, i);
	  {
	    i =  -4+(i);
	  }
	}
      }
    }
    return (true);
  case  -2:
    dprintf(">>>>ERROR<<<< "+("initloadFlc")+" on read only segment "+(seg)+"?\n");
    return false;
  default:
    return (true);
  }
}

/*  Called with SEG-LCK locked. */
/*  If you don't know what you are doing. DON'T DO IT! */
/*  Compute inverse hash function so that ent can still be found. */

public static void amnesiaEnt(Ent ent)
{
  ent_SetId(ent, hash2int(1+(ent_Seg(ent)), ent_Id(ent)));
  ent_SetDty(ent, null);
  ent_SetPus(ent, 0);
  if (a2b(ent_Blk(ent)))
    if (blk_Typ_P(ent_Blk(ent), dirTyp))
      blk_SetTyp(ent_Blk(ent), indTyp);
  ent_SetSeg(ent,  -1);
  ent_SetAge(ent, 0x80);
}


public static boolean blkFree(Ent ent)
{
LblkFree:while (true) {
  {
    int seg = ent_Seg(ent);
    if ((ent_Acc(ent))!=(accwrite))
      {
	dprintf(">>>>ERROR<<<< "+("blkFree")+": "+(ent_Seg(ent))+":"+(ent_Id(ent))+" without ACCWRITE\n");
	return false;
      }
    else {
      lck(seg_Lck(seg));
      if (!(initloadFlc_P(seg)))
	{
	  unlck(seg_Lck(seg));
	  return false;
	}
      else if ((seg_FlcLen(seg))>=((flcLen)-1))
	{
	  unlck(seg_Lck(seg));
	  flushFlc(seg, (flcLen)-2);
	  continue LblkFree;
	}
      else {
	seg_Flc(seg)[seg_FlcLen(seg)] = ent_Id(ent);
	seg_SetFlcLen(seg, (seg_FlcLen(seg))+1);
	amnesiaEnt(ent);
	unlck(seg_Lck(seg));
	return (true);
      }
    }
  }
}
}


public static int flcFill(int seg)
{
  byte []fstr = new byte[4];
  int flen = 0;
  int status;
BunlckAndReturn:while (true) {
    lck(seg_Lck(seg));
    if ((seg_FlcLen(seg))>=1)
      {
	unlck(seg_Lck(seg));
	return success;
      }
    else if (!(tryLck_P(seg_Fck(seg))))
      {
	unlck(seg_Lck(seg));
	dprintf("WARNING: "+("flcFill")+" failed -- branch never tried before! Segment "+(seg)+" "+(seg_Str(seg))+"\n");
	return retryerr;
      }
    else 
    flen = btNext(seg_FlHan(seg), noByts, 0, fstr);
    if (err_P(flen))
      {
	lck(emptyBlkLck);
	{
	  int xnum = (seg_Used(seg))+((flcLen)/2);
	  initLeafBlk(emptyBlk, xnum, indTyp);
	  if (a2b(blkio_FileExtend(seg_Port(seg), emptyBlk, seg_Bsiz(seg), xnum)))
	    {
	      if (a2b(ioDiag))
		dprintf(">>>>EXTENDING<<<< Segment "+(seg)+" "+(seg_Str(seg))+" by "+((flcLen)/2)+" blocks.\n");
	      {
		int i = 0;
		while (!((i)>((flcLen)/2))) {
		  seg_Flc(seg)[seg_FlcLen(seg)] = (xnum)-(i);
		  seg_SetFlcLen(seg, (seg_FlcLen(seg))+1);
		  seg_SetUsed(seg, (seg_Used(seg))+1);
		  {
		    i = (i)+1;
		  }
		}
	      }
	      {
		byte []usedStr = new byte[4];
		long2str(usedStr, 0, seg_Used(seg));
		btPut(seg_RtHan(seg), usedByts, 4, usedStr, 4);
	      }
	      unlck(emptyBlkLck);
	      {
		status = success;
		break BunlckAndReturn;
	      }
	    }
	  else {
	    dprintf(">>>>ERROR<<<< No more file space available! Segment "+(seg)+" "+(seg_Str(seg))+"\n");
	    unlck(emptyBlkLck);
	    {
	      status = noroom;
	      break BunlckAndReturn;
	    }
	  }
	}
      }
    else {
      unlck(seg_Lck(seg));
      {
	int []longAra = new int[(flcLen)+1];
	byte []xstr = new byte[0x100];
	int []respkt = new int[pktSize];
	int result = success;
	substringMove(fstr, 0, flen, xstr, 0);
	longAra[0] = 0;
	pkt_SetSkeyCount(respkt, 0);
	result = btScan(seg_FlHan(seg), remScan, xstr, flen, noByts, endOfChain, getMethod("wb.Ents", "flcProc", new Class[]{byte[].class, int.class, byte[].class, int.class, int[].class}), longAra, respkt, 1);
	if (((result)==(success)
	    || (result)==(notpres)
	    || (result)==(terminated)))
	  {
	    lck(seg_Lck(seg));
	    {
	      int i = longAra[0];
	      while (!((i)<=0)) {
		seg_Flc(seg)[seg_FlcLen(seg)] = longAra[i];
		seg_SetFlcLen(seg, (seg_FlcLen(seg))+1);
		{
		  i = (i)-1;
		}
	      }
	    }
	    {
	      status = success;
	      break BunlckAndReturn;
	    }
	  }
	else {
	  unlck(seg_Fck(seg));
	  return result;
	}
      }
    }
  }
  unlck(seg_Lck(seg));
  unlck(seg_Fck(seg));
  return status;
}


public static int flcProc(byte []keystr,int klen,byte []vstr,int vlen,int []longAra)
{
  {
    int ct = longAra[0];
    if ((ct)<((flcLen)/2))
      {
	int num = str2long(keystr, 0);
	ct = (ct)+1;
	longAra[0] = ct;
	longAra[ct] = num;
	return success;
      }
    else return terminated;
  }
}

/* create-new-blk-ent leaves you with ACCWRITE accmode to blk */

public static Ent createNewBlkEnt(int seg)
{
LcreateNewBlkEnt:while (true) {
  lck(seg_Lck(seg));
  if (!(initloadFlc_P(seg)))
    {
      unlck(seg_Lck(seg));
      return null;
    }
  else if ((seg_FlcLen(seg))<=0)
    {
      unlck(seg_Lck(seg));
      {
	int res = flcFill(seg);
	if (realerr_P(res))
	  return null;
	else continue LcreateNewBlkEnt;
      }
    }
  else {
    seg_SetFlcLen(seg, (seg_FlcLen(seg))-1);
    {
      int bnum = seg_Flc(seg)[seg_FlcLen(seg)];
      unlck(seg_Lck(seg));
      return getEnt(seg, bnum, accwrite);
    }
  }
}
}
/*  End of stuff to deal with the free-list-cache (FLC) */

/*  try-get-ent returns an entry with accmode or #f if blk is lcked.  When */
/*  you are done with the entry you need to release-ent!. */

public static Ent tryGetEnt(int seg,int blkNum,int accmode)
{
  {
    Ent buk = getBukWait(seg, blkNum);
    {
      Ent ent = buk;
Lentloop: while (true) {
      if (!(a2b(ent)))
	{
	  relBuk(seg, blkNum);
	  tgeFct = 1+(tgeFct);
	  return null;
	}
      else if (!((seg)==(ent_Seg(ent))
	      && (blkNum)==(ent_Id(ent))))
	{
	  ent = ent_Next(ent);
	  continue Lentloop;
	}
      else if ((blk_Id(ent_Blk(ent)))!=(blkNum))
	{
	  relBuk(seg, blkNum);
	  dprintf(">>>>ERROR<<<< corrupted buffer "+(ent_Seg(ent))+":"+(blk_Id(ent_Blk(ent)))+" <> "+(blkNum)+"\n");
	  tgeFct = 1+(tgeFct);
	  return null;
	}
      else if ((accnone)==(accmode))
	{
	  ent_SetRef(ent, 1+(ent_Ref(ent)));
	  relBuk(seg, blkNum);
	  tgeCt = 1+(tgeCt);
	  return ent;
	}
      else if ((accnone)==(ent_Acc(ent)))
	{
	  ent_SetAcc(ent, accmode);
	  ent_SetRef(ent, 1+(ent_Ref(ent)));
	  relBuk(seg, blkNum);
	  tgeCt = 1+(tgeCt);
	  return ent;
	}
      else {
	relBuk(seg, blkNum);
	tgeFct = 1+(tgeFct);
	return null;
      }
      }
    }
  }
}


public static Ent chainFindEnt(Han han,int accmode,byte []keyStr,int kLen,int []pkt)
{
  Ent ent = cacheEntEnable_P
      && a2b(han_Last(han))
    ?tryGetEnt(han_Seg(han), han_Last(han), accmode)
    :null;
  if (a2b(ent)
      && leaf_P(ent_Blk(ent))
      && (blk_TopId(ent_Blk(ent)))==(han_Id(han))
      && a2b(blk_FindPos(ent_Blk(ent), keyStr, kLen, pkt))
      && ((pkt_MatchType(pkt))==(match)
	     || ((pkt_MatchType(pkt))==(pastp)
		        || (pkt_MatchType(pkt))==(qpastp))
		    && (pkt_MatchPos(pkt))>(blkDataStart)))
    {
      tceCt = (tceCt)+1;
      return ent;
    }
  else {
    if (a2b(ent))
      releaseEnt(ent, accmode);
    tceFct = (tceFct)+1;
    ent = getEnt(han_Seg(han), han_Id(han), accnone);
    if ((!(root_P(ent_Blk(ent)))
        || blk_Typ_P(ent_Blk(ent), seqTyp)))
      {
	dprintf(">>>>ERROR<<<< "+("btOpen")+": not a B-tree root "+(ent_Seg(ent))+":"+(ent_Id(ent))+"\n");
	releaseEnt(ent, accnone);
	ent = null;
      }
    else ent = findEnt(ent, leaf,  -1, keyStr, kLen);
    if (!(a2b(ent)))
      ;
    else if ((accread)==(accmode))
      ;
    else if (ents_EntUpdateAccess_P(ent, accread, accmode))
      ;
    else {
      releaseEnt(ent, accread);
      ent = null;
    }
    if (a2b(ent))
      ent = chainFind(ent, accmode, keyStr, kLen, pkt);
    if (a2b(ent))
      han_SetLast(han, ent_Id(ent));
    return ent;
  }
}

/*  I havent put the call to PREV:PREV-K-ENT inside here, */
/*  as both paths need to call it - rjz */


public static Ent chainFindPrevEnt(Han han,int accmode,byte []keyStr,int kLen,int []pkt)
{
  Ent ent = cacheEntEnable_P
      && a2b(han_Last(han))
    ?tryGetEnt(han_Seg(han), han_Last(han), accmode)
    :null;
  if (a2b(ent)
      && leaf_P(ent_Blk(ent))
      && (blk_TopId(ent_Blk(ent)))==(han_Id(han))
      && a2b(blk_FindPos(ent_Blk(ent), keyStr, kLen, pkt))
      && ((pkt_MatchType(pkt))==(match)
	     || (pkt_MatchType(pkt))==(matchend)
	     || ((pkt_MatchType(pkt))==(pastp)
		        || (pkt_MatchType(pkt))==(qpastp))
		    && (pkt_MatchPos(pkt))>(blkDataStart)))
    {
      tceCt = (tceCt)+1;
      return ent;
    }
  else {
    if (a2b(ent))
      releaseEnt(ent, accmode);
    tceFct = (tceFct)+1;
    ent = prev_FindPrevEnt(getEnt(han_Seg(han), han_Id(han), accnone), leaf,  -1, keyStr, kLen);
    if (!(a2b(ent)))
      ;
    else if ((accmode)==(accread))
      ;
    else if (ents_EntUpdateAccess_P(ent, accread, accmode))
      ;
    else {
      releaseEnt(ent, accread);
      ent = null;
    }
    return ent;
  }
}

/* (REL-BUK! seg blk-num) */
/* (edprintf "all ents in use!\n") */


public static Ent getEnt(int seg,int blkNum,int accmode)
{
  if (0 > (blkNum))
    {
      dprintf(">>>>ERROR<<<< negative block number "+(blkNum)+"\n");
      return null;
    }
  else if ((blkNum)>=(seg_Used(seg)))
    {
      dprintf(">>>>ERROR<<<< bad block number "+(blkNum)+"\n");
      return null;
    }
  else {
    Ent ent = getBukWait(seg, blkNum);
Lentloop: while (true) {
    if (!(a2b(ent)))
      {
	ent = tryGetFreeEnt(seg, blkNum);
	if (a2b(ent))
	  {
	    ent_SetNext(ent, getBuk(seg, blkNum));
	    setBuk(seg, blkNum, ent);
	    ent_SetAcc(ent, accpend);
	    ent_SetSeg(ent, seg);
	    ent_SetId(ent, blkNum);
	    ent_SetAge(ent,  -127);
	    ent_SetDty(ent, null);
	    ent_SetPus(ent, 0);
	    ent_SetRef(ent, 1);
	    relBuk(seg, blkNum);
	    if ((accmode)==(accwrite))
	      {
		ent_SetAcc(ent, accwrite);
		ent_SetDty(ent, true);
		initLeafBlk(ent_Blk(ent), blkNum, indTyp);
		geCt = 1+(geCt);
		return ent;
	      }
	    else if (a2b(blkio_Read(seg_Port(seg), ent_Blk(ent), seg_Bsiz(seg), blkNum)))
	      {
		ent_SetAcc(ent, accmode);
		if ((blk_Id(ent_Blk(ent)))!=(blkNum))
		  dprintf(">>>>ERROR<<<< corrupted blk "+(ent_Seg(ent))+":"+(blkNum)+" <> "+(blk_Id(ent_Blk(ent)))+"\n");
		geCt = 1+(geCt);
		return ent;
	      }
	    else {
	      ent_SetRef(ent, 0);
	      ent_SetAcc(ent, accnone);
	      geFct = 1+(geFct);
	      return null;
	    }
	  }
	else {
	  ent = getBukWait(seg, blkNum);
	  continue Lentloop;
	}
      }
    else if (!((seg)==(ent_Seg(ent))
	    && (blkNum)==(ent_Id(ent))))
      {
	ent = ent_Next(ent);
	continue Lentloop;
      }
    else if ((blk_Id(ent_Blk(ent)))!=(blkNum))
      {
	relBuk(seg, blkNum);
	dprintf(">>>>ERROR<<<< corrupted buffer "+(ent_Seg(ent))+":"+(blk_Id(ent_Blk(ent)))+" <> "+(blkNum)+"\n");
	geFct = 1+(geFct);
	return null;
      }
    else if ((accnone)==(accmode))
      {
	ent_SetRef(ent, 1+(ent_Ref(ent)));
	relBuk(seg, blkNum);
	geCt = 1+(geCt);
	return ent;
      }
    else if ((accnone)==(ent_Acc(ent)))
      {
	ent_SetAcc(ent, accmode);
	ent_SetRef(ent, 1+(ent_Ref(ent)));
	relBuk(seg, blkNum);
	geCt = 1+(geCt);
	return ent;
      }
    else {
      relBuk(seg, blkNum);
      geFct = 1+(geFct);
      return null;
    }
    }
  }
}


public static Ent switchEnt(Ent oldEnt,int oldacc,int newNum,int newacc)
{
  {
    Ent newEnt = getEnt(ent_Seg(oldEnt), newNum, accnone);
    releaseEnt(oldEnt, oldacc);
    if (a2b(newEnt))
      ents_EntUpdateAccess_P(newEnt, accnone, newacc);
    return newEnt;
  }
}

/* minimum real NUM-ENTS-CT is 12*number of processes */
/* minimum NUM-BUKS IS 2, MAYBE 3 (due to how get-free-ent works) */
/* minimum real BLK-SIZE is 1.5K */


public static int initWb(int maxNumEntsCt,int maxNumBuks,int maxBlkSize)
{
  if (a2b(freeEntLck))
    {
      dprintf(">>>>ERROR<<<< "+("initWb")+": already initialized\n");
      return argerr;
    }
  else {
    clearStats();
    numBuks = maxNumBuks;
    blkSize = maxBlkSize;
    emptyBlk = new byte[blkSize];
    emptyBlkLck = makeLck( -3);
    freeBukCntr = 0;
    freeEntLck = makeLck( -1);
    flushEntCntr = 0;
    flushEntLck = makeLck( -2);
    bukTab = new Ent[numBuks];
    lckTab = new Lck[numBuks];
    entTab = new Ent[entTabInc];
    {
      int i = numBuks;
      while (!(0==(i))) {
	lckTab[(i)-1] = makeLck((i)-1);
	{
	  i = (i)-1;
	}
      }
    }
    {
      int seg = 9;
      while (!(0 > (seg))) {
	lck(seg_Lck(seg));
	seg_SetFlc(seg, new int [flcLen]);
	seg_SetFlcLen(seg, 0);
	unlck(seg_Lck(seg));
	{
	  seg = (seg)-1;
	}
      }
    }
    lck(freeEntLck);
    {
      int i = maxNumEntsCt;
      Ent bent = null;
      while (!(0==(i))) {
	{
	  Ent newent = ent_MakeEnt(numEntsCt);
	  if (a2b(newent))
	    {
	      ent_SetNext(newent, bent);
	      bent = newent;
	      entTab[numEntsCt] = newent;
	      ent_SetTag(newent, numEntsCt);
	      numEntsCt = 1+(numEntsCt);
	      if (0==((numEntsCt)%(entTabInc)))
		{
		  Ent []tmpEntTab = (Ent[]) resizeArray(entTab, (int)((entTabInc)+(numEntsCt)));
		  if (a2b(tmpEntTab))
		    {
		      entTab = tmpEntTab;
		    }
		  else i = 1;
		}
	    }
	  else i = 1;
	}
	{
	  i = (i)-1;
	}
      }
      freeEnts = bent;
    }
    unlck(freeEntLck);
    return numEntsCt;
  }
}


public static int finalWb()
{
  if (a2b(freeEntLck))
    {
      {
	int seg = (numSegs)-1;
	while (!(0 > (seg))) {
	  if (!(segFree_P(seg)))
	    closeSeg(seg, true);
	  {
	    seg = (seg)-1;
	  }
	}
      }
      lck(freeEntLck);
      {
	int i = numEntsCt;
	while (!(0==(i))) {
	  
	  entTab[ -1+(i)] = null;
	  numEntsCt =  -1+(numEntsCt);
	  {
	    i = (i)-1;
	  }
	}
      }
      unlck(freeEntLck);
      {
	int seg = (numSegs)-1;
	while (!(0 > (seg))) {
	  lck(seg_Lck(seg));
	  
	  seg_SetFlc(seg, null);
	  seg_SetFlcLen(seg, 0);
	  unlck(seg_Lck(seg));
	  {
	    seg = (seg)-1;
	  }
	}
      }
      {
	int i = numBuks;
	while (!(0==(i))) {
	  
	  lckTab[(i)-1] = null;
	  {
	    i = (i)-1;
	  }
	}
      }
      
      entTab = null;
      
      lckTab = null;
      
      bukTab = null;
      
      flushEntLck = null;
      
      freeEntLck = null;
      
      emptyBlk = null;
      
      emptyBlkLck = null;
      return success;
    }
  else return argerr;
}


public static void checkBlk(byte []blk)
{
  {
    int bEnd = blk_End(blk);
    {
      int bPos = blkDataStart;
Llp: while (true) {
      {
	int sPos = nextField(blk, 1+(bPos));
	if ((sPos)==(bEnd))
	  
	  return;
	else if ((sPos)<(bEnd))
	  {
	    {
	      bPos = nextCnvpair(blk, bPos);
	      continue Llp;
	    }
	  }
	else {
	  dprintf(">>>>ERROR<<<< "+("checkBlk")+": blk "+(blk_Id(blk))+" past end "+(sPos)+"\n");
	  
	  return;
	}
      }
      }
    }
  }
}


public static void checkKeyOrder(byte []blk)
{
  byte []splitStr = new byte[0x100];
  int spos = splitKeyPos(blk);
  if (a2b(spos))
    reconThisKey(blk, spos, splitStr, 0, 0x100);
}


public static int doSegBuffers(int seg,java.lang.reflect.Method func)
{
  {
    int i = numBuks;
    Ent ent = null;
Llp: while (true) {
    if (!(a2b(ent)))
      if (0==(i))
	return success;
      else {
	int T_i = (i)-1;
	ent = bukTab[(i)-1];
	i = T_i;
	continue Llp;
      }
    else if ((0 > (seg)
        || (seg)==(ent_Seg(ent))))
      {
	int ans = intFunInvoke(func, null, new Object[] {ent});
	if (success_P(ans))
	  {
	    ent = ent_Next(ent);
	    continue Llp;
	  }
	else return ans;
      }
    else {
      ent = ent_Next(ent);
      continue Llp;
    }
    }
  }
}


public static int checkBuffer(Ent ent)
{
  if (!(0==(ent_Ref(ent))))
    {
      dprintf(">>>>ERROR<<<< Entry still referenced: "+(ent_Seg(ent))+":"+(ent_Id(ent))+"\n");
      ent_SetRef(ent, 0);
    }
  if ((accnone)!=(ent_Acc(ent)))
    {
      dprintf(">>>>ERROR<<<< Entry still lcked: "+(ent_Seg(ent))+":"+(ent_Id(ent))+"\n");
      ents_EntUpdateAccess_P(ent, ent_Acc(ent), accnone);
    }
  return success;
}


public static void checkAccess()
{
  checkLcks();
  doSegBuffers( -1, getMethod("wb.Ents", "checkBuffer", new Class[]{Ent.class}));
}


/*  This routine needs to deal with lck issues. */
/*  TBD needs to give error if lcked. */

public static int flushBuffer(Ent ent)
{
  if ((accnone)!=(ent_Acc(ent)))
    return terminated;
  else if (ent_Dty_P(ent))
    if (a2b(ents_EntWrite(ent)))
      return success;
    else return retryerr;
  else return success;
}


public static int purgeBuffer(Ent ent)
{
  if (ent_Dty_P(ent))
    {
      if (((ent_Acc(ent))==(accwrite)
	  || (ent_Acc(ent))==(accpend)))
	dprintf("  Purging "+((ent_Acc(ent))==(accwrite)
	?"ACCWRITE"
	:"ACCPEND")+" entry: "+(ent_Seg(ent))+":"+(ent_Id(ent))+"\n");
      ents_EntWrite(ent);
    }
  amnesiaEnt(ent);
  return success;
}

}
