;; "blkio.scm" WB-tree File Based Associative String Data Base System.
;; Copyright (C) 1991, 1992, 1993, 2000, 2003 Free Software Foundation, Inc.
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place, Suite 330, Boston,
;; MA 02111, USA.
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of WB.
;;
;; The exception is that, if you link the WB library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the WB library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name WB.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; WB, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for WB, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

;;; routines in this file return pass/fail (#t/#f)

(require 'i/o-extensions)

(pragma.c "#include \"wbsys.h\""
	  "#ifdef unix"
	  "# include <sys/types.h>"
	  "# include <sys/stat.h>"
	  "# include <fcntl.h>"
	  "# include <unistd.h>"
	  "#endif")

(java:import "Ents" "Scan" "Prev" "Stats" "Blink" "Del" "Segs"
	     "Handle" "Wbdefs" "Wbsys" "Blk" "BooleanMethods" "Pkt"
	     "Ent" "Seg" "Db")

(define io-diag #f)

(define (blkio:init!) #f)

(define (blkio:create-file name bsiz)
  (open-output-file name))

(define (blkio:open-modify-file name bsiz)
  (open-io-file name))

(define (blkio:open-read-only-file name bsiz)
  (open-input-file name))

(define (blkio:file-close file) (close-port file))

(define (blkio:read file blk bsiz blknum)
  (file-position file (* bsiz blknum))
  (cond ((= bsiz (subbytes-read! blk 0 bsiz file))
	 (if io-diag (dprintf "rd:%*s%c %10ld\n"
			      (+ 1 (- (BLK:LEVEL blk) LEAF)) ""
			      (BLK:TYP blk) blknum))
	 (if (BLK:TYP? blk FRL-TYP)
	     (set! read-fl-ct (+ read-fl-ct 1))
	     (set! read-ct (+ read-ct 1)))
	 #t)
	(else (edprintf " couldn't read blk %ld\n" blknum)
	      #f)))

(define (blkio:write file blk bsiz blknum)
  (file-position file (* bsiz blknum))
  (cond ((= bsiz (subbytes-write blk 0 bsiz file))
	 (if io-diag (dprintf "wr:%*s%c %10ld\n"
			      (+ 1 (- (BLK:LEVEL blk) LEAF)) ""
			      (BLK:TYP blk) blknum))
	 (if (BLK:TYP? blk FRL-TYP)
	     (set! write-fl-ct (+ write-fl-ct 1))
	     (set! write-ct (+ write-ct 1)))
	 #t)
	(else (edprintf " couldn't write blk %ld\n" blknum)
	      #f)))

(define (blkio:file-extend file blk bsiz blknum)
  (file-position file (* bsiz blknum))
  (cond ((= bsiz (subbytes-write blk 0 bsiz file))
	 (if io-diag (dprintf "Extending file %d blks\n"
			      (+ (quotient FLC-LEN 2) 1)))
	 #t)
	(else (edprintf " couldn't extend file %ld\n" blknum)
	      #f)))
