;; "db.scm" WB-tree File Based Associative String Data Base System.
;; Copyright (C) 1991, 1992, 1993, 2000 Free Software Foundation, Inc.
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place, Suite 330, Boston,
;; MA 02111, USA.
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of WB.
;;
;; The exception is that, if you link the WB library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the WB library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name WB.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; WB, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for WB, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

;;; DEBUG AND TEST CODE

(define (wb:err? x) (and (number? x) (err? x)))

(define flush-ents ents-flush)

(define (create-bt seg typ wcb)
  (let* ((a-han (HAN:MAKE-HAN))
	 (ans (bt-create seg typ a-han wcb)))
    (if (err? ans) ans a-han)))

(define (open-bt seg blknum wcb)
  (let* ((a-han (HAN:MAKE-HAN))
	 (ans (bt-open seg blknum a-han wcb)))
    (if (err? ans) ans a-han)))

(define close-bt! bt-close)

;;; rem! removes key-str and value.  returns #t if found, #f if not.
(define (bt:rem! han key-str)
  (success? (bt-rem han key-str (string-length key-str) #f)))

(define (bt:rem* han key-str key2-str)
  (define tmpstr (make-bytes 256))
  (substring-move! key-str 0 (string-length key-str) tmpstr 0)
  (bt-rem-range han tmpstr (if (zero? (string-length key-str))
			       START-OF-CHAIN (string-length key-str))
		key2-str (if (zero? (string-length key2-str))
			     END-OF-CHAIN (string-length key2-str))))

;;; rem removes key-str and value.  returns value.
(define (bt:rem han key-str)
  (let* ((tmp-str (make-bytes 256))
	 (tlen (bt-rem han key-str (string-length key-str) tmp-str)))
    (if (err? tlen) #f (substring tmp-str 0 tlen))))

;;; put adds an key-str value pair to the database whose root is blk
(define (bt:put! han key-str val-str)
  (bt-put han
	  key-str (string-length key-str)
	  val-str (string-length val-str)))

(define (bt:put han key-str val-str)
  (success?
   (bt-write han
	     key-str (string-length key-str)
	     val-str (string-length val-str))))

;;; long-string
(define (db:put! han ckey val)
  (define len (string-length val))
  (define (put! han ckey val)
    (and (wb:err? (bt:put! han ckey val))
	 (slib:error 'db:put 'could-not-put han ckey val)))
  (cond ((> len 255)
	 (put! han ckey (substring val 0 255))
	 (do ((kdx 1 (+ 1 kdx))
	      (idx 510 (+ 255 idx))
	      (ldx 255 idx))
	     ((>= idx len)
	      (put! han (string-append ckey (bytes kdx))
		    (substring val ldx len)))
	   (put! han (string-append ckey (bytes kdx))
		 (substring val ldx idx))))
	(else (put! han ckey val))))

;;; long-string
(define (db:get han ckey)
  (define val (bt:get han ckey))
  (and val
       (case (string-length val)
	 ((255)
	  (let ((ckey+ (string-append ckey (bytes 255)))
		(strs (list val)))
	    (define bytes:0 (bytes 0))
	    (define tail strs)
	    (define (maproc k v)
	      (set-cdr! tail (list v))
	      (set! tail (cdr tail))
	      #t)
	    (do ((res (bt:scan han 0 (string-append ckey bytes:0) ckey+
			       maproc 1)
		      (bt:scan han 0 (caddr res) ckey+ maproc 1)))
		((not (= -1 (car res)))
		 (apply string-append strs)))))
	 (else val))))

;;; get returns a string of the value or #f
(define (bt:get han key)
  (let* ((tmp-str (make-bytes 256))
	 (tlen (bt-get han key (string-length key) tmp-str)))
    (if (err? tlen) #f (substring tmp-str 0 tlen))))

;;; next returns a string of the next key-str or #f if at end.
;;; (bt:next blk #f) or (bt:next blk "") returns the first key-str.
;;; to make BLINK happy I'm passing length 0 instead of START-OF-CHAIN

(define (bt:next han key-str)
  (let* ((tmp-str (make-bytes 256))
	 (tlen
	  (if (and key-str (> (string-length key-str) 0) )
	      (bt-next han key-str (string-length key-str) tmp-str)
	      (bt-next han "" START-OF-CHAIN tmp-str))))
    (if (err? tlen) #f (substring tmp-str 0 tlen))))

;;; prev returns a string of the previous key-str or #f if at end.
;;; (bt:prev blk #f) or (bt:prev blk "") returns the last key-str.

(define (bt:prev han key-str)
  (let* ((tmp-str (make-bytes 256))
	 (tlen
	  (if (and key-str (> (string-length key-str) 0))
	      (bt-prev han key-str (string-length key-str) tmp-str)
	      (bt-prev han "" END-OF-CHAIN tmp-str))))
    (if (err? tlen) #f (substring tmp-str 0 tlen))))

(define (create-db seg typ namestr)
  (if (char? typ) (set! typ (char->integer typ)))
  (let* ((tmp-str (make-bytes 256))
	 (a-han (create-bt seg typ 0))
	 (d-han (open-bt seg 1 (+ WCB-SAP WCB-SAR))))
    (if (or (err? a-han) (err? d-han)) #f
	(begin
	  (long2str! tmp-str 1 (HAN:ID a-han))
	  (byte-set! tmp-str 0 4)
	  (bt-put d-han namestr (string-length namestr) tmp-str 5)
	  (close-bt! d-han)
	  a-han))))

(define (open-db seg namestr)
  (let* ((tmp-str (make-bytes 256))
	 (d-han (open-bt seg 1 (+ WCB-SAP WCB-SAR))))
    (if (err? d-han) #f
	(let* ((tlen (bt-get d-han namestr (string-length namestr) tmp-str)))
	  (close-bt! d-han)
	  (if (err? tlen) #f
	      (if (eqv? tlen 5)
		  (open-bt seg (str2long tmp-str 1) 0)
		  #f))))))

(define (bt:scan bthan op kstr1 kstr2 scmproc blklim)
  (let ((ikey (make-bytes 256))
	(respkt (make-vector PKT-SIZE))
	(proc
	 (and scmproc
	      (lambda (key klen val vlen extra)
		(let ((res (scmproc (substring key 0 klen) (substring val 0 vlen))))
		  (cond ((number? res) res)
			((not res) NOTPRES)
			((boolean? res) SUCCESS)
			((not (string? res)) TYPERR)
			((substring-move! res 0 (string-length res) val 0)
			 (string-length res))))))))
    (PKT:SET-SKEY-COUNT! respkt 0)
    (PKT:SET-SKEY-LEN! respkt (string-length kstr1))
    (substring-move! kstr1 0 (string-length kstr1) ikey 0)
    (let ((res (bt-scan bthan op ikey (PKT:SKEY-LEN respkt)
			kstr2 (string-length kstr2) proc #f respkt blklim)))
      (list res (PKT:SKEY-COUNT respkt) (substring ikey 0 (PKT:SKEY-LEN respkt))))))

;;(trace pkt:skey-count bt:scan bt-scan) (set! *qp-width* 0)
