/* WB-tree File Based Associative String Data Base System.
 * Copyright (C) 1991, 1992, 1993, 2000, 2003 Free Software Foundation, Inc.
 * Copyright 2007 Clear Methods, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place, Suite 330, Boston, MA
 * 02111, USA.
 *
 * As a special exception, the Free Software Foundation gives permission
 * for additional uses of the text contained in its release of WB.
 *
 * The exception is that, if you link the WB library with other files
 * to produce an executable, this does not by itself cause the
 * resulting executable to be covered by the GNU General Public License.
 * Your use of that executable is in no way restricted on account of
 * linking the WB library code into it.
 *
 * This exception does not however invalidate any other reasons why
 * the executable file might be covered by the GNU General Public License.
 *
 * This exception applies only to the code released by the
 * Free Software Foundation under the name WB.  If you copy
 * code from other Free Software Foundation releases into a copy of
 * WB, as the General Public License permits, the exception does
 * not apply to the code that you add in this way.  To avoid misleading
 * anyone as to the status of such modified files, you must delete
 * this exception notice from them.
 *
 * If you write modifications of your own for WB, it is your choice
 * whether to permit this exception to apply to your modifications.
 * If you do not wish that, delete this exception notice.
 */
package wb;

import static wb.Wbsys.*;
import static wb.Blink.*;
import static wb.Segs.*;
import static wb.Handle.*;
import static wb.Blkio.*;
import static wb.Del.*;
import static wb.Ents.*;
import static wb.Prev.*;
import static wb.Scan.*;
import static wb.Stats.*;
import static wb.Wbdefs.*;
import static wb.BooleanMethods.*;
import static wb.Blk.*;
import static wb.Han.*;
import static wb.Pkt.*;

public class Db {

public static boolean wb_Err_P(int x)
{
    return 0 > x && x >= maxerr;
}

public static int flushEnts(int trynum, int flushnum)
{
    return entsFlush(trynum, flushnum);
}

public static Han createBt(int seg, int typ, int wcb)
{
    Han aHan = hanMakeHan();
    int ans = btCreate(seg, typ, aHan, wcb);
    if (err_P(ans)) return null;
    return aHan;
}

public static Han openBt(int seg, int blknum, int wcb)
{
    Han aHan = hanMakeHan();
    int ans = btOpen(seg, blknum, aHan, wcb);
    if (err_P(ans)) return null;
    return aHan;
}

public static void closeBt(Han han)
{
    btClose(han);
}

public static Han createDb(int seg, char typ, String nameStr)
{
    return createDb(seg, Character.getNumericValue(typ), nameStr);
}

public static Han createDb(int seg, int typ, String nameStr)
{
    byte[] tmpStr = new byte[0x100];
    Han aHan = createBt(seg, typ, 0);
    Han dHan = openBt(seg, 1, (wcbSap + wcbSar));
    if (! a2b(aHan) && ! a2b(dHan)) return null;
    else {
	long2str(tmpStr, 1, han_Id(aHan));
	tmpStr[0] = 4;
	//should the name of the db too be UTF-8 encoded?
	btPut(dHan, stringToBytes(nameStr), strlen(nameStr), tmpStr, 5);
	closeBt(dHan);
	return aHan;
    }
}

//		(define (open-db seg namestr)
//		  (let* ((tmp-str (make-bytes 256))
//			 (d-han (open-bt seg 1 (+ WCB-SAP WCB-SAR))))
//		    (if (err? d-han) #f
//			(let* ((tlen (bt-get d-han namestr (string-length namestr) tmp-str)))
//			  (close-bt! d-han)
//			  (if (err? tlen) #f
//			      (if (eqv? tlen 5)
//				  (open-bt seg (str2long tmp-str 1) 0)
//				  #f))))))

public static Han openDb(int seg, String nameStr)
{
    byte[] tmpStr = new byte[0x100];
    Han dHan = openBt(seg, 1, (wcbSap + wcbSar));
    if (dHan == null) return null;
    else {
	int tlen =
	    btGet(dHan, stringToBytes(nameStr), nameStr.length(), tmpStr);
	closeBt (dHan);
	if (err_P(tlen)) return null;
	if (tlen == 5) return openBt(seg, str2long(tmpStr, 1), 0);
	return null;
    }
}

/*  Del removes key-str and value.  returns #t if found, #f if not. */
//bt-rem!
public static boolean bt_Del(Han han, byte[] keyStr)
{
    return success_P(btRem(han, keyStr, strlen(keyStr), null));
}

// bt-rem*
public static int bt_Delete(Han han, byte[] keyStr, byte[] key2Str)
{
    byte []tmpstr = new byte[0x100];
    substringMove(keyStr, 0, strlen(keyStr), tmpstr, 0);
    return btRemRange(han,
		      tmpstr,
		      ! a2b(strlen(keyStr)) ? startOfChain : strlen(keyStr),
		      key2Str,
		      ! a2b(strlen(key2Str)) ? endOfChain : strlen(key2Str));
}

/*  rem removes key-str and value.  returns value. */
public static byte[] bt_Rem(Han han, byte[] keyStr)
{
    byte []tmpStr = new byte[0x100];
    int tlen = btRem(han, keyStr, strlen(keyStr), tmpStr);
    if(err_P(tlen)) return null;
    else return subArray(tmpStr, tlen);
}

/*  put adds an key-str value pair to the database whose root is blk */
//bt-put!
public static void bt_Put(Han han, byte[] keyStr, byte[] valStr)
{
    btPut(han, keyStr, strlen(keyStr), valStr, strlen(valStr));
}

//bt-put
public static boolean bt_Insert(Han han, byte[] keyStr, byte[] valStr)
{
    return success_P(btWrite(han,
			     keyStr, strlen(keyStr),
			     valStr, strlen(valStr)));
}

/*  get returns a string of the value or #f */
public static byte[] bt_Get(Han han, byte[] keyStr)
{
    byte []tmpStr = new byte[0x100];
    int tlen = btGet(han, keyStr, strlen(keyStr), tmpStr);
    if (err_P(tlen)) return null;
    else return subArray(tmpStr, tlen);
}

/*  next returns a string of the next key-str or #f if at end. */
/*  (bt:next blk #f) or (bt:next blk "") returns the first key-str. */
/*  to make BLINK happy I'm passing length 0 instead of START-OF-CHAIN */
public static byte[] bt_Next(Han han, byte[] keyStr)
{
    byte []tmpStr = new byte[0x100];
    int tlen = (a2b(keyStr) && strlen(keyStr)>0)
	? btNext(han, keyStr, strlen(keyStr), tmpStr)
	: btNext(han, noByts, startOfChain, tmpStr);
    if (err_P(tlen)) return null;
    else return subArray(tmpStr, tlen);
}

/*  prev returns a string of the previous key-str or #f if at end. */
/*  (bt:prev blk #f) or (bt:prev blk "") returns the last key-str. */
public static byte[] bt_Prev(Han han, byte[] keyStr)
{
    byte []tmpStr = new byte[0x100];
    int tlen = (a2b(keyStr) && strlen(keyStr)>0)
	? btPrev(han, keyStr, strlen(keyStr), tmpStr)
	: btPrev(han, noByts, endOfChain, tmpStr);
    if (err_P(tlen)) return null;
    else return subArray(tmpStr, tlen);
}

/* Wrapper functions for db access that take strings instead
   of byte arrays */
public static boolean bt_Del(Han han, String keyStr)
{
    return bt_Del(han, stringToBytes(keyStr));
}

public static void bt_Delete(Han han, String keyStr, String key2Str)
{
    bt_Delete(han, stringToBytes(keyStr), stringToBytes(key2Str));
}

public static String bt_Get(Han han, String keyStr)
{
    byte[] byts = bt_Get(han, stringToBytes(keyStr));
    return bytesToString(byts);
}

public static String bt_Next(Han han, String keyStr)
{
    byte[] byts = bt_Next(han, stringToBytes(keyStr));
    return bytesToString(byts);
}


public static String bt_Prev(Han han, String keyStr)
{
    byte[] byts =  bt_Prev(han, stringToBytes(keyStr));
    return bytesToString(byts);
}

public static void bt_Put(Han han, String keyStr, String key2Str)
{
    bt_Put(han, stringToBytes(keyStr), stringToBytes(key2Str));
}

public static String bt_Rem(Han han, String keyStr)
{
    byte[] byts =  bt_Rem(han, stringToBytes(keyStr));
    return bytesToString(byts);
}

//   (define (bt:scan bthan op kstr1 kstr2 scmproc blklim)
//   (let ((ikey (make-bytes 256))
//	(respkt (make-pkt))
//	(proc
//	 (and scmproc
//	      (lambda (key klen val vlen extra)
//		(let ((res (scmproc (substring key 0 klen) (substring val 0 vlen))))
//		  (cond ((number? res) res)
//			((not res) NOTPRES)
//			((boolean? res) SUCCESS)
//			((not (string? res)) TYPERR)
//			((substring-move! res 0 (string-length res) val 0)
//			 (string-length res))))))))
//     (PKT:SET-SKEY-COUNT! respkt 0)
//     (PKT:SET-SKEY-LEN! respkt (string-length kstr1))
//     (substring-move! kstr1 0 (string-length kstr1) ikey 0)
//     (let ((res (scan:bt-scan bthan op ikey (PKT:SKEY-LEN respkt)
//			     kstr2 (string-length kstr2) proc #f respkt blklim)))
//       (list res (PKT:SKEY-COUNT respkt) (substring ikey 0 (PKT:SKEY-LEN respkt))))))

//Incomplete translation.
// public static int[] bt_Scan(int bthan, int op, byte[] kstr1, byte[] kstr2, int scmproc, int blklim)
// {
//     byte[] ikey = new byte[0x100];
//     int[] resPkt = Pkt.makePkt();
//     //Should make this work
//     //proc = a2b(scmproc) && a2b(bt_ScanHelper(key, klen, vlen, extra));
//     pkt_SetSkeyCount (resPkt, 0);
//     pkt_SetSkeyLen(resPkt, strlen(kstr1));
//     Wbsys.substringMove(kstr1, 0, strlen(kstr1), ikey, 0);
//     //Should make this work
//     //int res = btScan(bthan, op, ikey, Pkt.skeyLen(resPkt),
//     //		 kstr2, strlen(kstr2), proc, 0, resPkt, blklim);
//     {
// 	int[] resList  = new int[3];
// 	//resList[0] = res;
// 	resList[1] = pkt_SkeyCount(resPkt);
// 	//resList[2] = Wbsys.substring(ikey, 0, Pkt.skeyLen(resPkt));
// 	return resList;
//     }
// }

// private static int bt_ScanHelper(int key, int klen, int val, int vlen, int extra, int scmproc)
// {
//     int res = 0;
//     return res;
// }

//public static void scmproc(ikey, klen, val, vlen, extra)

// public static byte[] int2bytes(int i)
// {
//     byte[] B = new byte[4];
//     long2str(B, 0, i);
//     return B;
// }

}
