/* WB-tree File Based Associative String Data Base System.
 * Copyright (C) 1991, 1992, 1993, 2000, 2003 Free Software Foundation, Inc.
 * Copyright 2007 Clear Methods, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place, Suite 330, Boston, MA
 * 02111, USA.
 *
 * As a special exception, the Free Software Foundation gives permission
 * for additional uses of the text contained in its release of WB.
 *
 * The exception is that, if you link the WB library with other files
 * to produce an executable, this does not by itself cause the
 * resulting executable to be covered by the GNU General Public License.
 * Your use of that executable is in no way restricted on account of
 * linking the WB library code into it.
 *
 * This exception does not however invalidate any other reasons why
 * the executable file might be covered by the GNU General Public License.
 *
 * This exception applies only to the code released by the
 * Free Software Foundation under the name WB.  If you copy
 * code from other Free Software Foundation releases into a copy of
 * WB, as the General Public License permits, the exception does
 * not apply to the code that you add in this way.  To avoid misleading
 * anyone as to the status of such modified files, you must delete
 * this exception notice from them.
 *
 * If you write modifications of your own for WB, it is your choice
 * whether to permit this exception to apply to your modifications.
 * If you do not wish that, delete this exception notice.
 */
package wb;

import java.io.RandomAccessFile;
import static wb.Wbdefs.*;
import static wb.Lck.*;
import static wb.Han.*;
import static wb.Ents.*;

public class Seg {
    /*  The SEG (segment)  methods*/
    public static RandomAccessFile  seg_Port(int seg)
    {
	return  segdTab[seg].PORT;
    }

    public static int  seg_Bsiz(int seg)
    {
	return  segdTab[seg].BSIZ;
    }

    public static int  seg_Used(int seg)
    {
	return  segdTab[seg].USED;
    }

    public static String  seg_Str(int seg)
    {
	return  segdTab[seg].STRN;
    }

    public static Han seg_RtHan(int seg)
    {
	return  segdTab[seg].RT_HAN;
    }

    public static Han seg_FlHan(int seg)
    {
	return  segdTab[seg].FL_HAN;
    }

    public static Lck  seg_Lck(int seg)
    {
	return  segdTab[seg].FLCK;
    }

    public static Lck  seg_Fck(int seg)
    {
	return  segdTab[seg].FFCK;
    }

    public static int  seg_FlcLen(int seg)
    {
	return  segdTab[seg].FLC_LEN;
    }

    public static int[]  seg_Flc(int seg)
    {
	return  segdTab[seg].FLC;
    }

    public static void  seg_SetPort(int seg, RandomAccessFile port)
    {
	segdTab[seg].PORT = port;
    }

    public static void  seg_SetBsiz(int seg, int bsiz)
    {
	segdTab[seg].BSIZ =  bsiz;
    }

    public static void  seg_SetUsed(int seg, int used)
    {
	segdTab[seg].USED =  used;
    }

    public static void  seg_SetStr(int seg, String str)
    {
	segdTab[seg].STRN =  str;
    }

    public static void  seg_SetFlcLen(int seg, int flcLen)
    {
	segdTab[seg].FLC_LEN = flcLen;
    }

    public static void  seg_SetFlc(int seg, int[] flc)
    {
	segdTab[seg].FLC = flc;
    }

    public static Seg newSeg(){
	Seg seg = new Seg();
	seg.RT_HAN = hanMakeHan();
	seg.FL_HAN = hanMakeHan();
	seg.FLCK = makeLck(0);
	seg.FFCK = makeLck(0);
	return seg;
    }

    //Defining segTab here tentatively
    /*
      Seg [] segdTab[num_segs] =
      {
      {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
      {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
      {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
      {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
      {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
      {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
      {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
      {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
      {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
      {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0}
      };
    */
    // Looks like the above structure is defined to initialize all the values
    // to null(or 0s).

    public static Seg[] segdTab = new Seg[numSegs];

    static {
        for(int i = 0; i < segdTab.length; i++) segdTab[i] = newSeg();
    }

    //	Segd members
    RandomAccessFile PORT;
    int BSIZ;
    int USED;
    String STRN;
    Han RT_HAN, FL_HAN;
    Lck FLCK ; //T - tentative, should look at locking
    Lck FFCK ; //T
    int FLC_LEN;
    int []FLC;
}
