/*  "stats.scm" WB-tree File Based Associative String Data Base System. */
/*  Copyright (C) 1991, 1992, 1993, 2000 Free Software Foundation, Inc. */
/*  */
/*  This program is free software; you can redistribute it and/or modify */
/*  it under the terms of the GNU General Public License as published by */
/*  the Free Software Foundation; either version 2, or (at your option) */
/*  any later version. */
/*  */
/*  This program is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/*  GNU General Public License for more details. */
/*  */
/*  You should have received a copy of the GNU General Public License */
/*  along with this software; see the file COPYING.  If not, write to */
/*  the Free Software Foundation, 59 Temple Place, Suite 330, Boston, */
/*  MA 02111, USA. */
/*  */
/*  As a special exception, the Free Software Foundation gives permission */
/*  for additional uses of the text contained in its release of WB. */
/*  */
/*  The exception is that, if you link the WB library with other files */
/*  to produce an executable, this does not by itself cause the */
/*  resulting executable to be covered by the GNU General Public License. */
/*  Your use of that executable is in no way restricted on account of */
/*  linking the WB library code into it. */
/*  */
/*  This exception does not however invalidate any other reasons why */
/*  the executable file might be covered by the GNU General Public License. */
/*  */
/*  This exception applies only to the code released by the */
/*  Free Software Foundation under the name WB.  If you copy */
/*  code from other Free Software Foundation releases into a copy of */
/*  WB, as the General Public License permits, the exception does */
/*  not apply to the code that you add in this way.  To avoid misleading */
/*  anyone as to the status of such modified files, you must delete */
/*  this exception notice from them. */
/*  */
/*  If you write modifications of your own for WB, it is your choice */
/*  whether to permit this exception to apply to your modifications. */
/*  If you do not wish that, delete this exception notice. */



package wb;

import static wb.Ents.*;
import static wb.Blkio.*;
import static wb.Prev.*;
import static wb.Scan.*;
import static wb.Blink.*;
import static wb.Del.*;
import static wb.Segs.*;
import static wb.Handle.*;
import static wb.Wbdefs.*;
import static wb.Wbsys.*;
import static wb.Blk.*;
import static wb.BooleanMethods.*;
import static wb.Pkt.*;
import static wb.Seg.*;
import static wb.Ent.*;
import static wb.Db.*;
public class Stats {

/*  PERFORMANCE STATISTICS */


public static int nextCt = 0;

public static int nextFct = 0;

public static int prevCt = 0;

public static int prevFct = 0;

public static int getCt = 0;

public static int getFct = 0;

public static int putCt = 0;

public static int putFct = 0;

public static int remCt = 0;

public static int remFct = 0;

public static int geCt = 0;

public static int geFct = 0;

public static int tgeCt = 0;

public static int tgeFct = 0;

public static int tceCt = 0;

public static int tceFct = 0;


public static int chainsToNext = 0;

public static int deferredInserts = 0;

public static int splitIndexInserts = 0;

public static int indexScrewCase = 0;

public static int blockSplits = 0;

public static int blockDeletes = 0;

public static int deferredDeletes = 0;

public static int dirDtyCt = 0;


public static int readCt = 0;

public static int writeCt = 0;

public static int readFlCt = 0;

public static int writeFlCt = 0;

public static int flushCt = 0;


public static int clearStats()
{
  nextCt = 0;
  nextFct = 0;
  prevCt = 0;
  prevFct = 0;
  getCt = 0;
  getFct = 0;
  putCt = 0;
  putFct = 0;
  remCt = 0;
  remFct = 0;
  geCt = 0;
  geFct = 0;
  tgeCt = 0;
  tgeFct = 0;
  tceCt = 0;
  tceFct = 0;
  chainsToNext = 0;
  deferredInserts = 0;
  splitIndexInserts = 0;
  indexScrewCase = 0;
  blockSplits = 0;
  blockDeletes = 0;
  deferredDeletes = 0;
  dirDtyCt = 0;
  readCt = 0;
  writeCt = 0;
  readFlCt = 0;
  writeFlCt = 0;
  flushCt = 0;
  return success;
}


public static int cstats()
{
  stats();
  return clearStats();
}


public static int stats()
{
  dprintf("\n");
  dprintf("OPERATIONS SUMMARY:\n");
  dprintf("            NEXT     PREV      GET      PUT      REM   GETENT   CACHE    CACHE\n");
  dprintf("                                                                AVAIL?   VALID?\n");
  dprintf("   succ "+(nextCt)+" "+(prevCt)+" "+(getCt)+" "+(putCt)
   +" "+(remCt)+" "+(geCt)+" "+(tgeCt)+" "+(tceCt)
   +"\n");
  dprintf("   fail "+(nextFct)+" "+(prevFct)+" "+(getFct)+" "+(putFct)
   +" "+(remFct)+" "+(geFct)+" "+(tgeFct)+" "+(tceFct)
   +"\n\n");
  dprintf("INTERNAL OPERATIONS SUMMARY:\n");
  dprintf("   chains-to-next = "+(chainsToNext)+"\n");
  dprintf("   "+(blockSplits)+" block splits, "+(deferredInserts)+" deferred parent updates, "+(deferredInserts)+" undone\n");
  dprintf("   "+(splitIndexInserts)+" split index inserts; "+(indexScrewCase)+" index-insert screw cases \n");
  dprintf("   "+(blockDeletes)+" block deletes; "+(deferredDeletes)+" deferred block deletes\n\n");
  dprintf("I/O SUMMARY: "+(readCt)+" READS, "+(writeCt)+" WRITES, "+(flushCt)+" FLUSH (of dirty bufs) calls.\n");
  dprintf("             "+(readFlCt)+" FREELIST READS, "+(writeFlCt)+" FREELIST WRITES.\n");
  dprintf("             "+(dirDtyCt)+" DIRS LEFT DTY.\n");
  {
    int ops = max((nextFct)+(prevFct)+(getFct)+(putFct)+(remFct)+(nextCt)+(prevCt)+(getCt)
     +(putCt)+(remCt), 1);
    dprintf("   AVG BLKS REFERENCED PER OPERATION: "+((0x64*((geCt)+(tceCt)))/(ops))+"%"+"\n");
    dprintf("   AVG DISK I/Os PER OPERATION: "+((0x64*((readCt)+(writeCt)+(readFlCt)+(writeFlCt)))/(ops))+"%"+"; ");
    dprintf("   WRITE/READ RATIO: "+((0x64*((writeCt)+(writeFlCt)))/(max(1, (readCt)+(readFlCt))))+"%"+"\n");
    if ((putCt)>0)
      dprintf("   AVG DISK WRITES PER PUT: "+((0x64*(writeCt))/(putCt))+"%"+"\n");
  }
  dprintf("\n");
  dprintf("MODES IN EFFECT:\n");
  dprintf(""+(numEntsCt)+" buffers [hashed over "+(numBuks)+" buckets]; blksize (max)="+(blkSize)+"; FLC-LEN="+(flcLen)
   +"\n");
  dprintf("defer-insert-updates?= "+(deferInsertUpdates_P
	?"#t"
	:"#f")+"; defer-block-deletes="+(del_DeferBlockDeletes_P
	?"#t"
	:"#f")+"\n");
  dprintf("cache-ent-enable?="+(cacheEntEnable_P
	?"#t"
	:"#f")+"; clever-cache-enable="+(a2b(cleverCacheEnable)
	?"#t"
	:"#f")+"\n");
  dprintf("\n");
  return success;
}


public static void showBuffer(Ent ent)
{
  dprintf("SEG = "+(ent_Seg(ent))+" ID= "+(ent_Id(ent))+" ACC= "+(ent_Acc(ent))+" REF= "+(ent_Ref(ent))
   +" DTY= "+(ent_Dty_P(ent))+" AGE= "+(ent_Age(ent))+"\n");
}


public static boolean bufVerbose_P = (true);


public static int showBuffer1(Ent ent)
{
  if ((bufVerbose_P
      || (ent_Seg(ent))> -1))
    {
      dprintf(" "+(ent_Seg(ent))+":"+(ent_Id(ent))+" "+(hash2int(ent_Seg(ent), ent_Id(ent)))+" "+(ent_Acc(ent))
   +" "+(ent_Ref(ent))+" "+0+" "+(ent_Dty_P(ent))+" "+(ent_Age(ent))
   +"");
      if ((ent_Seg(ent))> -1)
	{
	  dprintf(" "+((blk_Level(ent_Blk(ent)))-(leaf))+" "+(blk_Typ(ent_Blk(ent)))+"\n");
	}
      else dprintf("\n");
    }
  return success;
}


public static int showBuffers()
{
  dprintf("\n");
  dprintf(" SEG:ID        BUK      ACC    REF READERS DTY  AGE LEVEL TYPE\n");
  return doSegBuffers( -1, getMethod("wb.Stats", "showBuffer1", new Class[]{Ent.class}));
}


public static int sb()
{
  return showBuffers();
}

}
