/* WB-tree File Based Associative String Data Base System.
 * Copyright (C) 1991, 1992, 1993, 2000 Free Software Foundation, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place, Suite 330, Boston, MA
 * 02111, USA.
 *
 * As a special exception, the Free Software Foundation gives permission
 * for additional uses of the text contained in its release of WB.
 *
 * The exception is that, if you link the WB library with other files
 * to produce an executable, this does not by itself cause the
 * resulting executable to be covered by the GNU General Public License.
 * Your use of that executable is in no way restricted on account of
 * linking the WB library code into it.
 *
 * This exception does not however invalidate any other reasons why
 * the executable file might be covered by the GNU General Public License.
 *
 * This exception applies only to the code released by the
 * Free Software Foundation under the name WB.  If you copy
 * code from other Free Software Foundation releases into a copy of
 * WB, as the General Public License permits, the exception does
 * not apply to the code that you add in this way.  To avoid misleading
 * anyone as to the status of such modified files, you must delete
 * this exception notice from them.
 *
 * If you write modifications of your own for WB, it is your choice
 * whether to permit this exception to apply to your modifications.
 * If you do not wish that, delete this exception notice.
 */

#include "wbsys.h"
#include "wbdefs.h"

#ifdef PLAN9
int errno;
#else
# include <errno.h>
#endif

/* this is where all diagnostic and error messages will appear */
FILE *diagout;

#ifndef STDC_INCLUDES
unsigned char *memcpy(dest, src, len)
     unsigned char *dest, *src;
     int len;
{
  while(len--) dest[len]=src[len];
  return src;
}
# define LACK_MEMMOVE
#endif
#ifdef GNUDOS
# define LACK_MEMMOVE
#endif
#ifdef LACK_MEMMOVE
unsigned char *memmove(dest, src, len)
     unsigned char *dest, *src;
     int len;
{
  if (dest >= src)
    while(len--) dest[len]=src[len];
  else
    {
      int i=0;
      while(i<len)
	{
	  dest[i]=src[i];
	  i++;
	}
    }
  return src;
}
#endif

LCK *last_lck=0;

LCK *make_lck(name)
     int name;
{
  LCK *l = (LCK *)malloc(sizeof (LCK));
  if (!l) {
    dprintf((diagout, ">>>>ERROR<<<< could not allocate lck\n"));
    exit(errno);
  }
  l->NEXT = last_lck;
  l->FLG = 0;
  l->NAME = name;
  last_lck=l;
  return l;
}

/* #define try_lck_P(lk) (!((lk)->FLG--)) */
int try_lck_P(lk)
     LCK *lk;
{
  return !(lk->FLG--);
}

/* #define lck(lk) {if((lk)->FLG--)dprintf((diagout,">>>>ERROR<<<< spinning on lck %d\n,(lk)->NAME"));} */
void lck(lk)
     LCK *lk;
{
  if (lk->FLG--) dprintf((diagout,">>>>ERROR<<<< spinning on lck %d\n",(lk)->NAME));
}

/* #define unlck(lk) {if((lk)->FLG)(lk)->FLG=0;else dprintf((diagout,"unlcking unlck %d\n",(lk)->NAME));} */
void unlck(lk)
     LCK *lk;
{
  if (lk->FLG) lk->FLG=0;
  else dprintf((diagout,">>>>ERROR<<<< unlcking unlck %d\n",lk->NAME));
}

void check_lcks()
{
  LCK *ll = last_lck;
  while(ll) {
    if (ll->FLG) {
      if (ll->NAME < 0)
	switch (ll->NAME) {
	  case -1:dprintf((diagout,">>>>ERROR<<<< free-ent-lck left lcked\n"));
	  case -2:dprintf((diagout,">>>>ERROR<<<< flush-buk-lck left lcked\n"));
	  default:dprintf((diagout,">>>>ERROR<<<< unknown lck left lcked\n"));
	}
      else dprintf((diagout,">>>>ERROR<<<< lck %d left lcked\n", ll->NAME));
      ll->FLG = 0;
    }
    ll = ll->NEXT;
  }
}

SEGD segd_tab[num_segs]
  = {{0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
       {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
       {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
       {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
       {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
       {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
       {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
       {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
       {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0},
       {0,0,0,0,{0,0,0},{0,0,0},{0,0,0},{0,0,0},0,0}};

ENTRY *ent_make_ent(tag)
     int tag;
{
  ENTRY *ent = (ENTRY *)malloc(sizeof (ENTRY));
  if (!ent) {
    dprintf((diagout, "WARNING: could not allocate entry\n"));
    return 0;
  }
  ent->TAG = tag;
  ent->NEXT = 0;
  ent->SEG = -1;
  ent->ID = -1;
  ent->BLK = (unsigned char *)malloc(blk_size);
  if (!ent->BLK) {
    dprintf((diagout, "WARNING: could not allocate blk for entry\n"));
    free(ent);
    return 0;
  }
  ent->AGE = 0;
  ent->DTY = 0;
  ent->PUS = 0;
  ent->ACC = accnone;
  ent->REF = 0;
  return ent;
}

HAND *make_han()
{
  HAND *han;
  han = (HAND *)calloc(1,sizeof(HAND));
  if (!han) {
    dprintf((diagout, ">>>>ERROR<<<< could not allocate handle\n"));
    exit(errno);
  }
/*
  han_set_num(han, 0);
  han_set_seg(han, 0);
  han_set_typ(han, 0);
  han_set_last(han, 0);
*/
  return han;
}
